from ..module import Module
from ..parameter import Parameter

class Linear(Module):
    def __init__(self, input_dim, output_dim, bias=True):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.weight = Parameter(shape=[self.output_dim, self.input_dim])

        if bias:
            self.bias = Parameter(shape=[self.output_dim, 1])
        else:
            self.bias = None

    def forward(self, x):
        if self.bias:
            z = self.weight @ x + self.bias
        else:
            z = self.weight @ x

        return z

    def inner_repr(self):
        return f"input_dim={self.input_dim}, output_dim={self.output_dim}, " \
               f"bias={True if self.bias is not None else False}"


import math
import norch

class AddBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        return [gradient, gradient]
    
class AddBroadcastedBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        x, y = self.input
        grad_x = self._reshape_gradient(gradient, x.shape)
        grad_y = self._reshape_gradient(gradient, y.shape)

        return [grad_x, grad_y]
    
    def _reshape_gradient(self, gradient, shape):
        # Reduce gradient dimensions to match the target shape dimensions
        while len(gradient.shape) > len(shape):
            gradient = gradient.sum(axis=0)

        # Sum along axes where the target shape dimension is 1
        for i in range(len(shape)):
            if shape[i] == 1:
                gradient = gradient.sum(axis=i, keepdim=True)
        
        return gradient
    
class SubBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        return [gradient, -gradient]
    
class SubBroadcastedBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        x, y = self.input
        grad_x = self._reshape_gradient(gradient, x.shape)
        grad_y = self._reshape_gradient(gradient, y.shape)
        return [grad_x, -grad_y]
    
    def _reshape_gradient(self, gradient, shape):
        # Reduce gradient dimensions to match the target shape dimensions
        while len(gradient.shape) > len(shape):
            gradient = gradient.sum(axis=0)

        # Sum along axes where the target shape dimension is 1
        for i in range(len(shape)):
            if shape[i] == 1:
                gradient = gradient.sum(axis=i)
        return gradient
    
class ScalarMulBackward:
    def __init__(self, x, scalar):
        self.input = [x]
        self.scalar = scalar

    def backward(self, gradient):
        return [gradient * self.scalar]

class ElementwiseMulBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        x = self.input[0]
        y = self.input[1]
        return [y * gradient, x * gradient]
    
class MatmulBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        x, y = self.input
        
        if x.ndim != y.ndim: # broadcasted case
            aux = (gradient @ y.transpose(-1,-2))
            aux_sum = aux.sum(axis=0)
            return [aux_sum, x.transpose(-1,-2) @ gradient]
        else:
            return [gradient @ y.transpose(-1,-2), x.transpose(-1,-2) @ gradient]
        
"""class PowBackward:
    def __init__(self, x, power):
        self.input = [x]
        self.power = power

    def backward(self, gradient):
        return [(gradient * self.power) * (self.input[0]) ** (self.power - 1)]"""
    
class PowBackward:
    def __init__(self, base, exponent):
        self.input = [base, exponent]

    def backward(self, gradient):
        base, exponent = self.input[0], self.input[1]

        if isinstance(base, (int, float)):
            grad_base = gradient * (base ** (exponent - 1))
            grad_exponent = (gradient * base ** exponent) * math.log(base)

        else:
            grad_base = gradient * exponent * (base ** (exponent - 1))
            grad_exponent = (gradient * base ** exponent) * (base.log())

        return [grad_base, grad_exponent]

    
class LogBackward:
    def __init__(self, x):
        self.input = [x]

    def backward(self, gradient):

        grad_input = gradient / self.input[0]

        return [grad_input]
       
class SumBackward:
    def __init__(self, x, axis=None, keepdim=False):
        self.input = [x]
        self.axis = axis
        self.keepdim = keepdim

    def backward(self, gradient):
        input_shape = self.input[0].shape.copy()
        if self.axis == -1:
            # If axis is None, sum reduces the tensor to a scalar.
            grad_output = float(gradient[[0] * len(gradient.shape)]) * self.input[0].ones_like()
        else:

            if self.keepdim:
                input_shape = input_shape[:self.axis] + [1] + input_shape[self.axis+1:]
            else:
                input_shape = input_shape[:self.axis] + input_shape[self.axis+1:]

            # Broadcast the gradient to the input shape along the specified axis.
            grad_output_shape = list(input_shape)
            grad_output = gradient.reshape(grad_output_shape)
            grad_output = grad_output + self.input[0].zeros_like()
        
        return [grad_output]
    
class ReshapeBackward:
    def __init__(self, x):
        self.input = [x]

    def backward(self, gradient):
        return [gradient.reshape(self.input[0].shape)]
    
class TransposeBackward:
    def __init__(self, x, axis1, axis2):
        self.input = [x]
        self.axis1 = axis1
        self.axis2 = axis2

    def backward(self, gradient):
        return [gradient.transpose(self.axis2, self.axis1)]
    
class TBackward:
    def __init__(self, x):
        self.input = [x]

    def backward(self, gradient):
        return [gradient.T]

    
class DivisionBackward:
    def __init__(self, x, y):
        self.input = [x, y]

    def backward(self, gradient):
        x, y = self.input
        grad_x = gradient / y
        grad_y = -1 * gradient * (x / (y * y))

        return [grad_x, grad_y]
    
class SinBackward:
    def __init__(self, x):
        self.input = [x]

    def backward(self, gradient):
        x = self.input[0]
        return [gradient * x.cos()]
    
class CosBackward:
    def __init__(self, x):
        self.input = [x]

    def backward(self, gradient):
        x = self.input[0]
        return [-gradient * x.sin()]
    
class MaxBackward:
    def __init__(self, x, axis=None, keepdim=False):
        self.input = [x]
        self.axis = axis
        self.keepdim = keepdim

    def backward(self, gradient):
        input_shape = self.input[0].shape.copy()
        if self.axis == -1:
            max_value = self.input[0].max()
            mask = self.input[0].equal(max_value)

            grad_output = float(gradient[[0] * len(gradient.shape)]) * self.input[0].ones_like()

            grad_output = (grad_output * mask) / mask.sum()[0]

        else:

            if self.keepdim:
                input_shape = input_shape[:self.axis] + [1] + input_shape[self.axis+1:]
            else:
                input_shape = input_shape[:self.axis] + input_shape[self.axis+1:]

            # Broadcast the gradient to the input shape along the specified axis.
            grad_output_shape = list(input_shape)
                            
            grad_output = gradient.reshape(grad_output_shape)
            grad_output = grad_output + self.input[0].zeros_like()
            max_values = self.input[0].max(axis=self.axis, keepdim=True)
            mask = self.input[0].equal(max_values)

            grad_output = (grad_output * mask)
        
        return [grad_output]

    
class MinBackward:
    def __init__(self, x, axis=None, keepdim=False):
        self.input = [x]
        self.axis = axis
        self.keepdim = keepdim

    def backward(self, gradient):
        input_shape = self.input[0].shape.copy()
        if self.axis == -1:
            min_value = self.input[0].min()
            mask = self.input[0].equal(min_value)

            grad_output = float(gradient[[0] * len(gradient.shape)]) * self.input[0].ones_like()

            grad_output = (grad_output * mask) / mask.sum()[0]

        else:
            if self.keepdim:
                input_shape = input_shape[:self.axis] + [1] + input_shape[self.axis+1:]
            else:
                input_shape = input_shape[:self.axis] + input_shape[self.axis+1:]

            # Broadcast the gradient to the input shape along the specified axis.
            grad_output_shape = list(input_shape)
            grad_output = gradient.reshape(grad_output_shape)
            grad_output = grad_output + self.input[0].zeros_like()
            max_values = self.input[0].min(axis=self.axis, keepdim=True)
            mask = self.input[0].equal(max_values)

            grad_output = (grad_output * mask)

        return [grad_output]


class CrossEntropyLossBackward:
    def __init__(self, logits, targets):
        self.input = [logits, targets]
    
    def backward(self, gradient):
        logits, targets = self.input

        if logits.ndim == 1:
            softmax = norch.softmax(logits, dim=0)
            grad_logits = (softmax - targets)
                
        elif logits.ndim == 2:
            # batched 
                batch_size = logits.shape[0]
                softmax = norch.softmax(logits, dim=1)

                grad_logits = (softmax - targets) / batch_size

        return [grad_logits, None]  # targets do not have a gradient

class SigmoidBackward:
    def __init__(self, input):
        self.input = [input]

    def backward(self, gradient):
        sigmoid_x = self.input[0].sigmoid()
        grad_input = gradient * sigmoid_x * (1 - sigmoid_x)

        return [grad_input]




import numpy as np
from .batch import Batch


class Dataloader(object):

    def __init__(self, dataset, batch_size=32):
        self.dataset = dataset
        self.batch_size = batch_size

    def __iter__(self):
        starts = np.arange(0, len(self.dataset), self.batch_size)

        for start in starts:
            end = start + self.batch_size
            batch_size = min(end, len(self.dataset)) - start
            yield Batch(self.dataset[start:end], batch_size)

    def __len__(self):
        return len(self.dataset) // self.batch_size



import unittest
import norch
from norch.utils import utils_unittests as utils
import torch
import os

class TestNNModuleLoss(unittest.TestCase):

    def setUp(self):
        self.device = os.environ.get('device')
        if self.device is None or self.device != 'cuda':
            self.device = 'cpu'

        print(f"Running tests on: {self.device}")

    def test_mse_loss(self):
        """
        Test the MSELoss
        """
        loss_fn_norch = norch.nn.MSELoss()
        loss_fn_torch = torch.nn.MSELoss()

        # Test case 1: Predictions and labels are equal
        predictions_norch = norch.Tensor([[1.1, 2, 3, 4], [1.1, 2, 3, 4]]).to(self.device)
        labels_norch = norch.Tensor([[1.1, 2, 3, 4], [1.1, 2, 3, 3]]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)

        predictions_torch = torch.tensor([[1.1, 2, 3, 4], [1.1, 2, 3, 4]], device=self.device)
        labels_torch = torch.tensor([[1.1, 2, 3, 4], [1.1, 2, 3, 3]], device=self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)        
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))
        
        # Test case 2: Predictions and labels are different
        predictions_norch = norch.Tensor([1.1, 2, 3, 4]).to(self.device)
        labels_norch = norch.Tensor([4, 3, 2.1, 1]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)

        predictions_torch = torch.tensor([1.1, 2, 3, 4], device=self.device)
        labels_torch = torch.tensor([4, 3, 2.1, 1], device=self.device)
        
        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)        
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))

    def test_cross_entropy_loss(self):
        """
        Test the CrossEntropyLoss
        """
        loss_fn_norch = norch.nn.CrossEntropyLoss()
        loss_fn_torch = torch.nn.CrossEntropyLoss()

        # Test case 1: Single class, single sample
        predictions_norch = norch.Tensor([2.0, 1.0, 0.1]).to(self.device)
        labels_norch = norch.Tensor([0]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)

        loss_torch_result = utils.to_torch(loss_norch).to(self.device)

        predictions_torch = torch.tensor([2.0, 1.0, 0.1], device=self.device)
        labels_torch = torch.tensor(0, device=self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)

        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))

        # Test case 2: Multiple classes, multiple samples
        predictions_norch = norch.Tensor([[0.5, 1.5, 2.5], [1.0, 2.0, 3.0]]).to(self.device)
        labels_norch = norch.Tensor([2, 1]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)
        

        predictions_torch = torch.tensor([[0.5, 1.5, 2.5], [1.0, 2.0, 3.0]], device=self.device)
        labels_torch = torch.tensor([2, 1], device=self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))
        
        # Test case 3: Edge case - all predictions are zero
        predictions_norch = norch.Tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]).to(self.device)
        labels_norch = norch.Tensor([1, 2]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)
        
        predictions_torch = torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], device=self.device)
        labels_torch = torch.tensor([1, 2], device=self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))

        # Test case 4: Class probabilities instead of class index
        predictions_norch = norch.Tensor([0.5, 0.2, 0.1]).to(self.device)
        labels_norch = norch.Tensor([1., 0, 0]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)
        
        predictions_torch = torch.tensor([0.5, 0.2, 0.1])
        labels_torch = torch.tensor([1., 0, 0])

        predictions_torch.to(self.device)
        labels_torch.to(self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))

        # Test case 4: Batched class probabilities instead of class index
        predictions_norch = norch.Tensor([[0.5, 0.2, 0.1], [0.1, 0.5, 0.7]]).to(self.device)
        labels_norch = norch.Tensor([[1., 0, 0], [0, 1, 0]]).to(self.device)
        loss_norch = loss_fn_norch.forward(predictions_norch, labels_norch)
        loss_torch_result = utils.to_torch(loss_norch).to(self.device)

        predictions_torch = torch.tensor([[0.5, 0.2, 0.1], [0.1, 0.5, 0.7]], device=self.device)
        labels_torch = torch.tensor([[1., 0, 0], [0, 1, 0]], device=self.device)

        loss_torch_expected = loss_fn_torch(predictions_torch, labels_torch)
        
        self.assertTrue(utils.compare_torch(loss_torch_result, loss_torch_expected))


class TestNNModuleActivationFn(unittest.TestCase):

    def setUp(self):
        self.device = os.environ.get('device')
        if self.device is None or self.device != 'cuda':
            self.device = 'cpu'

    def test_sigmoid_activation(self):
        """
        Test Sigmoid activation function
        """
        sigmoid_fn_norch = norch.nn.Sigmoid()
        sigmoid_fn_torch = torch.nn.Sigmoid()

        # Test case 1: Positive input
        x = norch.Tensor([[1, 2, 3]]).to(self.device)
        sigmoid_norch = sigmoid_fn_norch.forward(x)
        sigmoid_torch_result = utils.to_torch(sigmoid_norch).to(self.device)

        x = torch.tensor([[1, 2, 3]], device=self.device)
        
        sigmoid_torch_expected = sigmoid_fn_torch.forward(x)

        self.assertTrue(utils.compare_torch(sigmoid_torch_result, sigmoid_torch_expected))

        # Test case 1: Negative input
        x = norch.Tensor([-1, 2, -3]).to(self.device)
        sigmoid_norch = sigmoid_fn_norch.forward(x)
        sigmoid_torch_result = utils.to_torch(sigmoid_norch).to(self.device)

        x = torch.tensor([-1, 2, -3], device=self.device)
        
        sigmoid_torch_expected = sigmoid_fn_torch.forward(x)

        self.assertTrue(utils.compare_torch(sigmoid_torch_result, sigmoid_torch_expected))

        # Test case 1: Zero input
        x = norch.Tensor([0, 0, 0]).to(self.device)
        sigmoid_norch = sigmoid_fn_norch.forward(x)
        sigmoid_torch_result = utils.to_torch(sigmoid_norch).to(self.device)

        x = torch.tensor([0, 0, 0], device=self.device)

        sigmoid_torch_expected = sigmoid_fn_torch.forward(x)

        self.assertTrue(utils.compare_torch(sigmoid_torch_result, sigmoid_torch_expected))

    def test_softmax_activation(self):
        """
        Test Softmax activation function
        """    

        # Test different axes
        axes = [0, 1, 2, -1]

        # Define the input tensors for different test cases
        test_cases = [
            (norch.Tensor([[[1., 2, 3], [4, 5, 6]]]).to(self.device), torch.tensor([[[1., 2, 3], [4, 5, 6]]], device=self.device)),
            (norch.Tensor([[[1., -1, 0], [2, -2, 0]]]).to(self.device), torch.tensor([[[1., -1, 0], [2, -2, 0]]], device=self.device)),
            (norch.Tensor([[[0., 0, 0], [0, 0, 0]]]).to(self.device), torch.tensor([[[0., 0, 0], [0, 0, 0]]], device=self.device))
        ]

        for dim in axes:
            softmax_fn_norch = norch.nn.Softmax(dim=dim)
            softmax_fn_torch = torch.nn.Softmax(dim=dim)

            for norch_input, torch_input in test_cases:
                
                # Forward pass using norch
                softmax_norch = softmax_fn_norch.forward(norch_input)
                softmax_torch_result = utils.to_torch(softmax_norch).to(self.device)

                # Forward pass using torch
                softmax_torch_expected = softmax_fn_torch.forward(torch_input)

                # Compare the results
                self.assertTrue(utils.compare_torch(softmax_torch_result, softmax_torch_expected))




import setuptools
from setuptools.command.install import install
import subprocess

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

apt_dependencies = [
    'nvidia-cuda-toolkit'
]

for package in apt_dependencies:
    subprocess.check_call(['sudo', 'apt', 'install', package])

class CustomInstall(install):
    def run(self):
        subprocess.call(['make', '-C', 'build'])
        install.run(self)


setuptools.setup(
    name = "norch",
    version = "0.0.4",
    author = "Lucas de Lima",
    author_email = "nogueiralucasdelima@gmail.com",
    description = "A deep learning framework",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/lucasdelimanogueira/PyNorch",
    project_urls = {
        "Bug Tracker": "https://github.com/lucasdelimanogueira/PyNorch/issues",
        "Repository": "https://github.com/lucasdelimanogueira/PyNorch"
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages = setuptools.find_packages(),
    package_data={'norch': ['csrc/*', 'libtensor.so']},
    cmdclass={
        'install': CustomInstall,
    },
    python_requires = ">=3.6"
)
